/*
    Project: TCTEMP
    Version: V1.11
    File   : na_tc.c

    Copyright (c) 2006-2008 Author of TCTEMP. All rights reserved.
    Public Key Fingerprint: 75EB 6BC2 01B7 F6E7 4BD7 CC58 4A5F C393 19EE 6E69

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCTEMP binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "na.h"
#include "na_tc.h"
#include <ntdddisk.h>

/*
 * The interface to the TrueCrypt device driver is taken from the source
 * code of TrueCrypt 4.3 (and minor changes from TrueCrypt 5.1a)
 */

#define TC_MAX_PASSWORD_LEN         64
#define TC_MAX_PASSWORD             (TC_MAX_PASSWORD_LEN + 1)
#define TC_MIN_DRIVER_VERSION       0x0510
#define TC_MAX_DRIVER_VERSION       0x051a
#define TC_MAX_PATH                 260
#define TC_IOCTL(idx)               (0x222000 | ((idx) << 2))
#define TC_IOCTL_GET_DRIVER_VERSION TC_IOCTL(1)
#define TC_IOCTL_MOUNT              TC_IOCTL(3)
#define TC_IOCTL_DISMOUNT           TC_IOCTL(4)

#define TC_IOCTL_LEGACY(idx)        (0x72000 | ((idx) << 2))
#define TC_IOCTL_LEGACY_GET_DRIVER_VERSION TC_IOCTL_LEGACY(6)

typedef struct
{
    INT    nPasswordLen;
    CHAR   aszPassword[TC_MAX_PASSWORD];
} TC_PASSWORD, *PTC_PASSWORD;

#pragma pack(push, 1)

typedef struct
{
    INT         iResult;
    WCHAR       wszVolumeName[TC_MAX_PATH];
    TC_PASSWORD pwVolume;
    BOOL        fCache;
    INT         nDosDriveNo;
    INT         nBytesPerSector;
    BOOL        fMountReadOnly;
    BOOL        fMountRemovable;
    BOOL        fExclusiveAccess;
    BOOL        fMountManager;
    BOOL        fUserContext;
    BOOL        fPreserveTimestamp;
    BOOL        fPartitionInInactiveSysEncScope;
    INT         nPartitionInInactiveSysEncScopeDriveNo;
    BOOL        fProtectHiddenVolume;
    TC_PASSWORD pwHiddenVolume;
} TC_MOUNT_DATA, *PTC_MOUNT_DATA;

typedef struct
{
    INT    nDosDriveNo;
    BOOL   fIgnoreOpenFiles;
    INT    iResult;
} TC_DISMOUNT_DATA, *PTC_DISMOUNT_DATA;

#pragma pack(pop)

static BOOL TCAPI NaGetHostSectorSize(LPCWSTR pszVolumeName,
  PINT piBytesPerSector)
{
    WCHAR         wszNtDriveName[8];
    HANDLE        hDrive;
    DISK_GEOMETRY dg;
    ULONG         ul;
    BOOL          fResult;

    NaAssert(piBytesPerSector);

    *piBytesPerSector = 0;
    if (pszVolumeName[1] != L':') {
        /*
         * return 0 bytes per sector for device hosted volumes
         * (the TrueCrypt driver overwrites the iBytesPerSector
         * field for device hosted volumes anyway).
         */
        return pszVolumeName[0] == '\\';
    }

    if (!NaGetNtDriveName(wszNtDriveName, countof(wszNtDriveName),
      pszVolumeName[0])) {
        return FALSE;
    }

    if (!NaOpenFile(&hDrive, wszNtDriveName, GENERIC_READ,
      FILE_SHARE_READ|FILE_SHARE_WRITE, FILE_SYNCHRONOUS_IO_NONALERT)) {
        return FALSE;
    }

    fResult = FALSE;
    if (NaDeviceIoControl(hDrive, IOCTL_DISK_GET_DRIVE_GEOMETRY,
      NULL, 0, &dg, sizeof(dg), &ul) && ul == sizeof(dg)) {
        *piBytesPerSector = dg.BytesPerSector;
        fResult = TRUE;
    }

    NaCloseFile(hDrive);

    return fResult;
}

BOOL TCAPI NaTcOpen(PTC_CONTEXT pCtx)
{
    DWORD dwVersion;

    if (!NaOpenDevice(&pCtx->hDevice, L"\\Device\\TrueCrypt")) {
        return FALSE;
    }

    dwVersion = 0;
    if (!NaDeviceIoControl(pCtx->hDevice, TC_IOCTL_GET_DRIVER_VERSION,
      &dwVersion, sizeof(dwVersion), &dwVersion, sizeof(dwVersion), NULL)) {
        if (!NaDeviceIoControl(pCtx->hDevice, TC_IOCTL_LEGACY_GET_DRIVER_VERSION,
          &dwVersion, sizeof(dwVersion), &dwVersion, sizeof(dwVersion), NULL)) {
            NaDbgPrintf(L"Cannot query TrueCrypt driver version!");
            NaCloseFile(pCtx->hDevice);
            return FALSE;
        }
    }
    if (dwVersion < TC_MIN_DRIVER_VERSION ||
      dwVersion > TC_MAX_DRIVER_VERSION) {
        NaDbgPrintf(L"Unsupported TrueCrypt version (%x)!", (unsigned)dwVersion);
        NaCloseFile(pCtx->hDevice);
        return FALSE;
    }

    return TRUE;
}

void TCAPI NaTcClose(PTC_CONTEXT pCtx)
{
    NaCloseFile(pCtx->hDevice);
}

BOOL TCAPI NaTcMountVolume(PTC_CONTEXT pCtx, WCHAR chDriveLetter,
  LPCWSTR pszVolumeName, LPCSTR pPassword, DWORD dwPasswordLength)
{
    PTC_MOUNT_DATA ptmd = (PTC_MOUNT_DATA)pCtx->rgbBuffer;

    NaAssert(pszVolumeName);
    NaAssert(sizeof(TC_MOUNT_DATA) <= TC_BUFFER_SIZE);

    if (wcslen(pszVolumeName) >= countof(ptmd->wszVolumeName)) {
        NaDbgPrintf(L"TrueCrypt volume name is too long!");
        return FALSE;
    }
    if (dwPasswordLength) {
        NaAssert(pPassword);
        if (dwPasswordLength > TC_MAX_PASSWORD_LEN) {
            NaDbgPrintf(L"TrueCrypt password is too long!");
            return FALSE;
        }
    }

    RtlZeroMemory(ptmd, sizeof(*ptmd));
    wcscpy(ptmd->wszVolumeName, pszVolumeName);
    if (dwPasswordLength) {
        ptmd->pwVolume.nPasswordLen = dwPasswordLength;
        memcpy(ptmd->pwVolume.aszPassword, pPassword, dwPasswordLength);
    }
    if (!NaGetHostSectorSize(pszVolumeName, &ptmd->nBytesPerSector)) {
        NaDbgPrintf(L"Cannot query sector size of container host!");
        return FALSE;
    }
    ptmd->nDosDriveNo = towupper(chDriveLetter) - L'A';
    ptmd->fExclusiveAccess = TRUE;
    ptmd->fPreserveTimestamp = TRUE;
    if (!NaDeviceIoControl(pCtx->hDevice, TC_IOCTL_MOUNT, ptmd, sizeof(*ptmd),
      ptmd, sizeof(*ptmd), NULL)) {
        NaDbgPrintf(L"Cannot mount TrueCrypt volume!");
        return FALSE;
    } else if (ptmd->iResult != 0) {
        NaDbgPrintf(L"Cannot mount TrueCrypt volume (Error code 0x%x)!",
          ptmd->iResult);
        return FALSE;
    }

    return TRUE;
}

BOOL TCAPI NaTcDismountVolume(PTC_CONTEXT pCtx, WCHAR chDriveLetter)
{
    PTC_DISMOUNT_DATA ptdd = (PTC_DISMOUNT_DATA)pCtx->rgbBuffer;

    NaAssert(sizeof(TC_DISMOUNT_DATA) <= TC_BUFFER_SIZE);

    RtlZeroMemory(ptdd, sizeof(*ptdd));
    ptdd->nDosDriveNo = towupper(chDriveLetter) - L'A';
    ptdd->fIgnoreOpenFiles = TRUE;

    if (!NaDeviceIoControl(pCtx->hDevice, TC_IOCTL_DISMOUNT, ptdd,
      sizeof(*ptdd), ptdd, sizeof(*ptdd), NULL)) {
        NaDbgPrintf(L"Cannot dismount TrueCrypt volume!");
        return FALSE;
    } else if (ptdd->iResult != 0) {
        NaDbgPrintf(L"Cannot dismount TrueCrypt volume (Error code 0x%x)!",
          ptdd->iResult);
        return FALSE;
    }

    return TRUE;
}
